﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Diagnostics;

using Microsoft.Kinect;
using Microsoft.Samples.Kinect.WpfViewers;
using Microsoft.Kinect.Toolkit;
using Controllers;
using Postures;
using Gestures;


namespace KinectProject
{

    /// <summary>
    /// 
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            
            //KinectSensorManager 
            this.KinectSensorManager = new KinectSensorManager();
            this.KinectSensorManager.KinectSensorChanged += this.KinectSensorChanged;
            this.DataContext = this.KinectSensorManager;

            InitializeComponent();

            //KinectSensorChooser
            this.kinectSensorChooserUI.KinectSensorChooser = sensorChooser;
            sensorChooser.Start();

            // Bind the KinectSensor from the sensorChooser to the KinectSensor on the KinectSensorManager            
            Binding kinectSensorBinding = new Binding("Kinect");
            kinectSensorBinding.Source = this.sensorChooser;
            BindingOperations.SetBinding(this.KinectSensorManager, KinectSensorManager.KinectSensorProperty, kinectSensorBinding);

        }

        private void KinectSensorChanged(object sender, KinectSensorManagerEventArgs<KinectSensor> args)
        {
            if (null != args.OldValue)
            {
                this.UninitializeKinectServices(args.OldValue);
            }

            if (null != args.NewValue)
            {
                this.InitializeKinectServices(this.KinectSensorManager, args.NewValue);

            }
        }

        // Kinect enabled apps should customize which Kinect services and initialize here.
        private void InitializeKinectServices(KinectSensorManager kinectSensorManager, KinectSensor sensor)
        {
            // Application should enable all streams first.
            kinectSensorManager.ColorFormat = ColorImageFormat.RgbResolution640x480Fps30;
            kinectSensorManager.ColorStreamEnabled = true;

            sensor.SkeletonFrameReady += this.SkeletonsReady;
            kinectSensorManager.TransformSmoothParameters = new TransformSmoothParameters
            {
                Smoothing = 0.5f,
                Correction = 0.5f,
                Prediction = 0.5f,
                JitterRadius = 0.05f,
                MaxDeviationRadius = 0.04f
            };
            kinectSensorManager.SkeletonStreamEnabled = true;
            kinectSensorManager.KinectSensorEnabled = true;


            //create posture detector and add event handler
            postureDetector = new PostureDetector();
            postureDetector.PostureDetected += PostureOutput;
            
            //add postures to the detector
            AddPostures();

            //create keyboard controller
            keyboardController = new KeyboardController();

        }
        // Kinect enabled apps should uninitialize all Kinect services that were initialized in InitializeKinectServices() here.
        private void UninitializeKinectServices(KinectSensor sensor)
        {
            sensor.SkeletonFrameReady -= this.SkeletonsReady;

            if (postureDetector != null)
            {
                postureDetector.PostureDetected -= PostureOutput;
            }
            CarController.ClosePort();
        }

        private void SkeletonsReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            using (SkeletonFrame skeletonFrame = e.OpenSkeletonFrame())
            {
                if (skeletonFrame != null)
                {
                    if ((this.skeletonData == null) || (this.skeletonData.Length != skeletonFrame.SkeletonArrayLength))
                    {
                        this.skeletonData = new Skeleton[skeletonFrame.SkeletonArrayLength];
                    }

                    skeletonFrame.CopySkeletonDataTo(this.skeletonData);

                    //Get skeleton for processing
                    Skeleton currentSkeleton = (from s in this.skeletonData
                                                where s.TrackingState == SkeletonTrackingState.Tracked
                                                select s).FirstOrDefault();

                    if (currentSkeleton != null)
                    {
                        
                        switch (Controller.SelectedIndex)
                        {
                            case 2:
                                // Controller.SelectedIndex == 2 -> Car Controller
                                //pass the skeleton to postureDetector for processing
                                postureDetector.DetectPosture(currentSkeleton);

                                break;

                            default:
                                Debug.WriteLine("Controller -- Default Case Reached!");
                                break;
                        }                        
                    }
                }
            }
        }

        private void PostureOutput(string postureName)
        {
            if (Controller.SelectedIndex == 2) //car controller is selected
            {
                ControlCar(postureName);
            }
        }

        private void Controller_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            AddPostures();

            if (Controller.SelectedIndex == 2)  //CarController selected
            {
                if (comboBox.SelectedIndex == 0)
                {
                    MessageBox.Show("Please select a Port Name!", "Information", MessageBoxButton.OK, MessageBoxImage.Information);
                    return;
                }
                OpenComPort(comboBox.SelectedIndex);
            }
        }


        private void comboBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (Controller.SelectedIndex == 2 && comboBox.SelectedIndex != 0) //only perfrom when car controller is selected and com port name provided
            {                                
                OpenComPort(Controller.SelectedIndex);                
            }
        }

        private void OpenComPort(int portNo)
        {
            CarController.ClosePort();  //close existing port
            if (!CarController.OpenPort(comboBox.SelectedIndex))
            {
                MessageBox.Show("COM" + comboBox.SelectedIndex + " is not ready", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                comboBox.SelectedIndex = 0;
                return;
            }
            CarController.AllPowerOff(); //clear all signal after then port successfully open
        }

    }
}
